/*
 * main.c
 *
 * Created: 30.03.2013 17:37:23
 *  Author: Lupin
 *
 * ATTiny25 clocked with 8 MHz
 * Startup time should be short, so that bypass transistors will be turned on before ZXLD1350 starts operation
 */ 

#define F_CPU 8000000

// AVR includes
#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/sleep.h>
#include <util/atomic.h>

/*
 Macros for peter danneggers senskey function
*/
struct bits {                           // Macro to access bits like variables
	uint8_t b0:1;
	uint8_t b1:1;
	uint8_t b2:1;
	uint8_t b3:1;
	uint8_t b4:1;
	uint8_t b5:1;
	uint8_t b6:1;
	uint8_t b7:1;
} __attribute__((__packed__));
#define SBIT(port,pin) ((*(volatile struct bits*)&port).b##pin)


/****** Port definitions *******/
#define LED_PORT		PORTB
#define RED_PIN			(1<<PB0)
#define GREEN_PIN		(1<<PB4)
#define BLUE_PIN		(1<<PB1)
#define ADJ_PIN			(1<<PB5)

// senskey stuff by peter Dannegger
#define SENSKEY_A0      SBIT( PORTB, 3 )        // connect to senspad
#define SENSKEY_A0_DDR  SBIT( DDRB,  3 )

#define SENSKEY_B0      SBIT( PORTB, 2 )        // connect to cap to A0
#define SENSKEY_B0_DDR  SBIT( DDRB,  2 )
#define SENSKEY_B0_PIN  SBIT( PINB,  2 )

// Threshold to detect if object was moved close or removed from touchpad
// This sets the touchpad sensitivity
#define TP_THRESHOLD	8


// Touchpad macro definitions (main loop)
#define TP_UP			0
#define TP_DOWN			1
#define TP_HOLD			2

// HSV color space stuff
typedef uint16_t hue_t;
volatile hue_t h=0;
#define HUE_FBITS		13 // How many fraction bits for hue? 3 bits are needed for integer part
#define HUE_FMASK		((hue_t)(((hue_t)1<<HUE_FBITS)-1))
#define HUE_MAX			(((hue_t)6)<<HUE_FBITS)

// Speed of LED fading
uint8_t led_on=0;

// Counter for mainloop to read a new touchpad value (at 5Hz)
// Will be reset by the main loop
volatile uint8_t smplctr;



/*
senskey function by peter dannegger:
http://www.mikrocontroller.net/topic/156809#1483874

A0      B0      B0_PIN
0       0               full discharge
loop:
Z       Z               gap
Z       1               partial charge over pin A0 and senspad
Z       Z               gab
0       Z               partial charge over pin B0 only
0       Z       1       end
*/
#define MAX_CYCLE       2000
uint16_t senskey( void )
{
	uint16_t i = MAX_CYCLE + 1;

	SENSKEY_A0 = 0;
	SENSKEY_A0_DDR = 1;
	SENSKEY_B0 = 0;
	SENSKEY_B0_DDR = 1;           // discharge cap

	// Making this block atomic will distort the soft-PWM output
	ATOMIC_BLOCK(ATOMIC_FORCEON){
		do{
			SENSKEY_A0_DDR = 0;       // A0 = tristate
			SENSKEY_B0 = 1;           // B0 = weak high
			SENSKEY_B0_DDR = 1;       // B0 = strong high
			SENSKEY_B0_DDR = 0;       // B0 = weak high
			SENSKEY_B0 = 0;           // B0 = tristate
			SENSKEY_A0_DDR = 1;       // A0 = strong low
			if( --i == 0 )
			break;                          // timeout
		}while( SENSKEY_B0_PIN == 0 );      // until charged
	}
	SENSKEY_B0_DDR = 1;           // discharge cap
	return MAX_CYCLE - i;         // big value = key touched or timeout
}


/*
 * Converts HUE of HSV colors space into RGB color space
 * Saturation and Value component of HSV color space are always "1"
 * This function only moves along the outer circle of the HSV cone
 */
void hsv_to_rgb(void)	
{
  uint8_t f;
  
  f = (uint8_t)((h&HUE_FMASK)>>(HUE_FBITS-8));

  switch(h>>HUE_FBITS) {
	case 0:
	default:
		OCR0A = 255;     OCR1B = f;       OCR1A = 0;       break;
	case 1:
		OCR0A = 255 - f; OCR1B = 255;     OCR1A = 0;       break;
	case 2:
		OCR0A = 0;       OCR1B = 255;     OCR1A = f;       break;
	case 3:
		OCR0A = 0;       OCR1B = 255 - f; OCR1A = 255;     break;			
	case 4:
		OCR0A = f;       OCR1B = 0;       OCR1A = 255;     break;
	case 5:
		OCR0A = 255;     OCR1B = 0;       OCR1A = 255 - f; break;
  }
}


/*
 * Soft PWM done by timer1
 * Also serves as a timebase for the mainloop, updates the hue and
 * calculates new RGB values when the PWM period is finished
 */
SIGNAL(TIMER0_OVF_vect) {
	// Counter for main loop
	smplctr++;
	
	if(led_on) {
		// Calculate new PWM values
		hsv_to_rgb();
		// Increase hue
		if(h++>=HUE_MAX) h-=HUE_MAX;
		// ZXLD ADJ pin open -> Constant current ON
		DDRB &= ~ADJ_PIN;
	} else {
		// ZXLD ADJ pin low -> Constant current OFF
		DDRB |= ADJ_PIN;
		// Output transistors should be OFF constantly
		OCR0A = 255; OCR1B = 255; OCR1A = 255;
	}
	
}

/*
 * Simple moving average filter
 */
// Filter order (must be 2^n)
#define FILORD		8	// This is actually the length of the filter... order is FILORD-1
#define FILSHIFT	3   // log2(FILORD) - C macro for this...?!

// Moving average lowpass filter
int16_t lpfilter(int16_t sample) {
	static int16_t z[FILORD];
	static int8_t ctr;
	int8_t i;
	int16_t fil;
	
	z[ctr&(FILORD-1)] = sample;
	ctr++; fil=0;
	for(i=0 ; i<FILORD ; i++) {
		fil += z[i];
	}
	return fil>>FILSHIFT;
}

// Settle the lowpass filter to some sampled value
void lpfilter_settle(int16_t sample) {
	int8_t i;
	for(i=0 ; i<FILORD ; i++) lpfilter(sample);
}


int	main(void)
{
	// Filtered and current touchpad value
	int16_t tp_filter, tp_current;
	// Assume the touch area is not touched at the moment
	int8_t tp_state = TP_UP;
	uint8_t i=0;
	
	// Set ports to output - this will actively disable the ZXLD current source
	DDRB = RED_PIN | GREEN_PIN | BLUE_PIN | ADJ_PIN;
	
	// Disable USI and ADC clocks to reduce power
	PRR = (1<<PRUSI) | (1<<PRADC);
	// Disable Analog comparator to reduce power
	ACSR = (1<<ACD);
	
	
	// Set up PWM
	// Timer 0 and timer1 are used to make the PWM
	// Setup Timer0
	TCCR0A = (1<<COM0A0) | (1<<COM0A1) | (1<<WGM00) | (1<<WGM01); // OC0A set on compare match (LED OFF), count from 0x00 to 0xFF (Fast PWM mode), overflow interrupt on 0xFF
	TCCR0B = (1<<CS02);	// 122 Hz PWM
	// Setup Timer1
	OCR1C  = 0xFF;   // TOP value of Timer1 - reset Timer1 to 0x00 at 0xFF (count from 0 to 255)
	TCCR1 = (1<<PWM1A) | (1<<COM1A1) | (1<<COM1A0) | (1<<CS13) | (1<<CS10); // OC1A set on compare match (LED OFF), presc. 8 MHz / 256 = 31250 Hz... /256 = 122Hz PWM
	GTCCR = (1<<PWM1B) | (1<<COM1B1) | (1<<COM1B0) | (1<<PSR1) | (1<<PSR0); // OC1B set on compare match (LED OFF), PSR (Pre-Scale-Reset for Timer1 and Timer0 so they will be in sync)
	// Enable Timer0 overflow interrupt
	TIMSK = (1<<TOIE0);  // Timer 0 overflow interrupt - in this interrupt new values are loaded for the PWM and serves as timebase for main loop	
	
	
	// Initialize the lowpass filter
	tp_filter = tp_current = senskey();
	lpfilter_settle(tp_current);
	

	// Use IDLE mode for power saving because CLK_IO is still needed to operate the timer modules (to wake up from reset)
	set_sleep_mode(SLEEP_MODE_IDLE);
	// Activate timer IRQ
	sei();

	// Main loop
	while(1) {
		
		// Go to idle mode until next timer event
		sleep_mode();
			
		// Read touchpad and filter signals at 10Hz
		if(smplctr >= 12) {
				
			// Read touchpad value
			tp_current=0;
			for(i=0;i<8;i++) tp_current += senskey();
			tp_current >>= 3;
			
			// Sense a touch event by comparing the current value
			// to the lowpass filtered touchpad value
			
			// Check if the touchpad is pressed
			// If the touchpad already was pressed before then this event can only be triggered by
			// applying more force to the sensor area or directly touching the sensor area (do nothing if this happens)
			if(tp_current < (tp_filter - TP_THRESHOLD) && tp_state != TP_DOWN) {
				// "Button down event"
				
				// New touchpad state
				tp_state = TP_DOWN;
				
				// Toggle the lamp on/off
				led_on ^= 1;
				
				// Settle the lowpass filter to the current touchpad value
				// so that a touchpad change will be sensed relative to this value
				lpfilter_settle(tp_current);
				
			} else {
				// New touchpad state
				tp_state = TP_UP;
			}
			
			// Filter touchpad value
			tp_filter = lpfilter(tp_current);	
			
			// Start counting the timer from 0
			smplctr = 0;
		}

	};
	
	return 0;
}